from typing import Iterator

from suds import sudsobject

from oda_wd_client.base.api import WorkdayClient
from oda_wd_client.base.tools import suds_to_dict
from oda_wd_client.service.financial_management.types import (
    AccountingJournalData,
    Company,
    ConversionRate,
    ConversionRateType,
    CostCenterWorktag,
    Currency,
    ProjectWorktag,
    SpendCategory,
)
from oda_wd_client.service.financial_management.utils import (
    pydantic_accounting_journal_to_workday,
    pydantic_conversion_rate_to_workday,
    workday_company_to_pydantic,
    workday_conversion_rate_to_pydantic,
    workday_conversion_rate_type_to_pydantic,
    workday_cost_center_to_pydantic,
    workday_currency_to_pydantic,
    workday_project_to_pydantic,
    workday_spend_category_to_pydantic,
    workday_tax_applicability_to_pydantic,
)


class FinancialManagement(WorkdayClient):
    service = "Financial_Management"

    def get_currency_rates(
        self, return_suds_object=False
    ) -> Iterator[sudsobject.Object | ConversionRate]:
        method = "Get_Currency_Conversion_Rates"
        results = self._get_paginated(method, "Currency_Conversion_Rate")
        for rate in results:
            yield rate if return_suds_object else workday_conversion_rate_to_pydantic(
                suds_to_dict(rate)
            )

    def get_currency_rate_types(
        self, return_suds_object=False
    ) -> Iterator[sudsobject.Object | ConversionRateType]:
        method = "Get_Currency_Rate_Types"
        results = self._get_paginated(method, "Currency_Rate_Type")
        for rate_type in results:
            yield rate_type if return_suds_object else workday_conversion_rate_type_to_pydantic(
                suds_to_dict(rate_type)
            )

    def put_currency_rate(self, rate: ConversionRate) -> sudsobject.Object:
        data_object = pydantic_conversion_rate_to_workday(rate, client=self)
        return self._request(
            "Put_Currency_Conversion_Rate", Currency_Conversion_Rate_Data=data_object
        )

    def get_cost_centers(
        self, return_suds_object=False
    ) -> Iterator[sudsobject.Object | CostCenterWorktag]:
        method = "Get_Cost_Centers"
        results = self._get_paginated(method, "Cost_Center")
        for cost_center in results:
            yield cost_center if return_suds_object else workday_cost_center_to_pydantic(
                suds_to_dict(cost_center)
            )

    def get_companies(
        self, return_suds_object=False
    ) -> Iterator[sudsobject.Object | Company]:
        method = "Get_Workday_Companies"
        results = self._get_paginated(method, "Company")
        for company in results:
            yield company if return_suds_object else workday_company_to_pydantic(
                suds_to_dict(company)
            )

    def get_currencies(
        self, return_suds_object=False
    ) -> Iterator[sudsobject.Object | Currency]:
        method = "GetAll_Currencies"
        response = self._request(method)
        results = response.Currency_Data
        for currency in results:
            yield currency if return_suds_object else workday_currency_to_pydantic(
                suds_to_dict(currency)
            )

    def get_projects(
        self, return_suds_object: bool = False
    ) -> Iterator[sudsobject.Object | ProjectWorktag]:
        method = "Get_Basic_Projects"
        results = self._get_paginated(method, "Basic_Project")
        for project in results:
            yield project if return_suds_object else workday_project_to_pydantic(
                suds_to_dict(project)
            )

    def get_spend_categories(
        self, return_suds_object: bool = False
    ) -> Iterator[sudsobject.Object | SpendCategory]:
        method = "Get_Resource_Categories"
        results = self._get_paginated(method, "Resource_Category")
        for category in results:
            yield category if return_suds_object else workday_spend_category_to_pydantic(
                suds_to_dict(category)
            )

    def get_tax_applicabilities(
        self, return_suds_object: bool = False
    ) -> Iterator[sudsobject.Object | dict]:
        method = "Get_Tax_Applicabilities"
        results = self._get_paginated(method, "Tax_Applicability")
        for tax_applicability in results:
            if return_suds_object:
                yield tax_applicability
            else:
                yield workday_tax_applicability_to_pydantic(
                    suds_to_dict(tax_applicability)
                )

    def submit_accounting_journal(
        self, journal: AccountingJournalData
    ) -> sudsobject.Object:
        data_object = pydantic_accounting_journal_to_workday(journal, client=self)
        return self._request(
            "Submit_Accounting_Journal", Accounting_Journal_Data=data_object
        )



from datetime import date
from decimal import Decimal
from typing import Literal

from pydantic import BaseModel, Field, validator

from oda_wd_client.base.types import File, WorkdayReferenceBaseModel
from oda_wd_client.base.utils import parse_workday_date
from oda_wd_client.service.financial_management.types import (
    Company,
    CostCenterWorktag,
    Currency,
    ProjectWorktag,
    SpendCategory,
)

# All public imports should be done through oda_wd_client.types.resource_management
__all__: list = []


class TaxApplicability(WorkdayReferenceBaseModel):
    """
    Reference: https://community.workday.com/sites/default/files/file-hosting/productionapi/Resource_Management/v40.2/Submit_Supplier_Invoice.html#Tax_ApplicabilityObjectType  # noqa
    """

    _class_name = "Tax_ApplicabilityObject"
    workday_id: str
    workday_id_type: Literal["Tax_Applicability_ID"] = "Tax_Applicability_ID"
    # Code is human-readable text but not critical, so we default to empty string
    code: str = ""
    taxable: bool = True


class TaxOption(WorkdayReferenceBaseModel):
    """
    Reference: https://community.workday.com/sites/default/files/file-hosting/productionapi/Resource_Management/v40.2/Submit_Supplier_Invoice.html#Tax_OptionObjectType  # noqa
    """

    _class_name = "Tax_OptionObject"
    workday_id: str
    workday_id_type: Literal["Tax_Option_ID"] = "Tax_Option_ID"


class TaxCode(WorkdayReferenceBaseModel):
    """
    Reference: https://community.workday.com/sites/default/files/file-hosting/productionapi/Resource_Management/v40.2/Submit_Supplier_Invoice.html#Tax_CodeObjectType  # noqa
    """

    _class_name = "Tax_CodeObject"
    workday_id: str
    workday_id_type: Literal["Tax_Code_ID"] = "Tax_Code_ID"


class Supplier(WorkdayReferenceBaseModel):
    """
    Reference: https://community.workday.com/sites/default/files/file-hosting/productionapi/Resource_Management/v40.2/Get_Suppliers.html#SupplierType  # noqa
    """

    _class_name = "SupplierObject"
    workday_id: str
    workday_id_type: Literal["Supplier_ID"] = "Supplier_ID"
    reference_id: str | None
    name: str | None
    payment_terms: str | None
    address: str | None
    phone: str | None
    email: str | None
    url: str | None
    currency: str | None
    bank_account: str | None
    iban: str | None

    # Tax ID format and type varies by country. This is organization number in Norway.
    # Norway - VAT
    tax_id_no: str | None
    # Austria - UID
    tax_id_au: str | None
    # Belgium - NOTVA
    tax_id_be: str | None
    # Switzerland - EID
    tax_id_ch: str | None
    # Germany - USTIDNR
    tax_id_de: str | None
    # Denmark - MOMS
    tax_id_dk: str | None
    # Spain - IVA
    tax_id_es: str | None
    # Finland - ALV
    tax_id_fi: str | None
    # Great Britain - VATREGNO
    tax_id_gb: str | None
    # Ireland - VATNO (I think -- it's called "IRE-VATNO" in WD, but IRE is not a valid countrycode)
    tax_id_ir: str | None
    # Netherlands - BTWNR
    tax_id_nl: str | None
    # Poland - VATNIP
    tax_id_pl: str | None
    # Sweden - MOMSNR
    tax_id_se: str | None
    # USA - EIN
    tax_id_us: str | None


class TaxRate(WorkdayReferenceBaseModel):
    """
    Reference: https://community.workday.com/sites/default/files/file-hosting/productionapi/Resource_Management/v40.2/Submit_Supplier_Invoice.html#Tax_RateObjectType  # noqa
    """

    _class_name = "Tax_RateObject"
    workday_id: str
    workday_id_type: Literal["Tax_Rate_ID"] = "Tax_Rate_ID"


class TaxRecoverability(WorkdayReferenceBaseModel):
    """
    Reference: https://community.workday.com/sites/default/files/file-hosting/productionapi/Resource_Management/v40.2/Submit_Supplier_Invoice.html#Tax_RecoverabilityObjectType  # noqa
    """

    _class_name = "Tax_RecoverabilityObject"
    workday_id: str
    workday_id_type: Literal[
        "Tax_Recoverability_Object_ID"
    ] = "Tax_Recoverability_Object_ID"


class TaxRateOptionsData(BaseModel):

    """
    Reference: https://community.workday.com/sites/default/files/file-hosting/productionapi/Resource_Management/v40.2/Submit_Supplier_Invoice.html#Tax_Rate_Options_DataType  # noqa

    With some (in)sane defaults
    """

    tax_rate: TaxRate
    tax_recoverability: TaxRecoverability = TaxRecoverability(
        workday_id="Fully_Recoverable"
    )
    tax_option: TaxOption = TaxOption(workday_id="CALC_TAX_DUE")


class FinancialAttachmentData(File):
    """
    Reference: https://community.workday.com/sites/default/files/file-hosting/productionapi/Resource_Management/v40.2/Submit_Supplier_Invoice.html#Financials_Attachment_DataType  # noqa
    """

    field_type = "Financials_Attachment_DataType"


class SupplierInvoiceLine(BaseModel):
    """
    Reference: https://community.workday.com/sites/default/files/file-hosting/productionapi/Resource_Management/v40.2/Submit_Supplier_Invoice.html#Supplier_Invoice_Line_Replacement_DataType  # noqa
    """

    order: int | None
    description: str = "-No description-"
    tax_rate_options_data: TaxRateOptionsData | None
    tax_applicability: TaxApplicability | None
    tax_code: TaxCode | None
    spend_category: SpendCategory | None
    cost_center: CostCenterWorktag | None
    project: ProjectWorktag | None
    gross_amount: Decimal = Field(max_digits=18, decimal_places=3)
    tax_amount: Decimal | None = Field(max_digits=18, decimal_places=3)
    total_amount: Decimal | None = Field(max_digits=18, decimal_places=3)
    budget_date: date | None


class SupplierInvoice(WorkdayReferenceBaseModel):
    """
    Reference: https://community.workday.com/sites/default/files/file-hosting/productionapi/Resource_Management/v40.2/Submit_Supplier_Invoice.html#Supplier_Invoice_DataType  # noqa
    """

    workday_id_type: Literal[
        "Supplier_invoice_Reference_ID"
    ] = "Supplier_invoice_Reference_ID"
    invoice_number: str
    company: Company
    currency: Currency
    supplier: Supplier
    invoice_date: date
    due_date: date
    total_amount: Decimal = Field(max_digits=26, decimal_places=6)
    tax_amount: Decimal = Field(max_digits=26, decimal_places=6)
    tax_option: TaxOption | None

    lines: list[SupplierInvoiceLine]
    attachments: list[FinancialAttachmentData] | None

    _normalize_dates = validator("invoice_date", "due_date", allow_reuse=True)(
        parse_workday_date
    )



from oda_wd_client.service.resource_management.types import (
    CostCenterWorktag,
    FinancialAttachmentData,
    SpendCategory,
    Supplier,
    SupplierInvoice,
    SupplierInvoiceLine,
    TaxApplicability,
    TaxCode,
    TaxOption,
    TaxRate,
    TaxRateOptionsData,
    TaxRecoverability,
)

__all__ = [
    "CostCenterWorktag",
    "SpendCategory",
    "Supplier",
    "SupplierInvoice",
    "SupplierInvoiceLine",
    "TaxApplicability",
    "TaxCode",
    "TaxOption",
    "TaxRate",
    "TaxRateOptionsData",
    "TaxRecoverability",
    "FinancialAttachmentData",
]



from base64 import b64encode
from typing import Self

import magic
from pydantic import BaseModel
from suds import sudsobject

from oda_wd_client.base.api import WorkdayClient
from oda_wd_client.base.utils import get_id_from_list

mime = magic.Magic(mime=True)


class WorkdayReferenceBaseModel(BaseModel):
    """
    Base class for all Workday reference models.
    A reference model is a model that is used to reference another pre-existing object in Workday.
    These models are used to have a simple/convenient way to generate reference objects for Workday, through the use of
    workday_id and workday_id_type.
    """

    workday_id: str | None
    workday_id_type: str
    workday_parent_id: str | None = None
    workday_parent_type: str | None = None

    # This is the name of the class in Workday. Usually ends with "Object" (i.e. "SupplierObject")
    _class_name: str | None = None

    def wd_object(
        self,
        client: WorkdayClient,
        class_name: str | None = None,
    ) -> sudsobject.Object:
        class_name = class_name or self._class_name
        assert (
            class_name
        ), "WD Class name must be supplied on class or call to wd_object"

        ref_obj = client.factory(f"ns0:{class_name}Type")
        id_obj = client.factory(f"ns0:{class_name}IDType")
        id_obj.value = self.workday_id
        id_obj._type = self.workday_id_type
        if self.workday_parent_id:
            id_obj._parent_id = self.workday_parent_id
            id_obj._parent_type = self.workday_parent_type

        ref_obj.ID.append(id_obj)
        return ref_obj

    @classmethod
    def from_id_list(cls, id_list: list, **extra) -> Self | None:
        """
        Make an instance of Self based on an ID list, using self.workday_id_type to look up the specified Workday ID

        The ID lists used for references are usually in the following format:

        id_list = {
            "Field_Reference": {
                "ID": [
                    {"_type": "Field_ID", value="ABC"},
                    {"_type": "WID", value="123456789"}
                ]
            }
        }

        `get_id_from_list(id_list, "Field_ID")` would give us "ABC".

        As the "field ID" value is defined on Self, as `workday_id_type`, we can simplify this lookup, and directly
        instantiate an instance based on the class definition and ID list.

        Example use:
            > SpendCategory.from_id_list(data["Spend_Category_Reference"]["ID"])

        :param id_list: List of IDs
        :param extra: Arbitrary kwargs are passed onto `cls.__init__()`
        :return: instance of Self
        """
        workday_id = get_id_from_list(
            id_list, cls.__fields__["workday_id_type"].default
        )
        if workday_id:
            return cls(workday_id=workday_id, **extra)
        return None


class File(BaseModel):
    """
    Representing a file in Workday, complete with MIME-identification and serialization

    Used as part of various fields within the web service
    """

    # All documents need a field type which is unique for each implementation (i.e. "Financials_Attachment_DataType"
    # or "Worker_Document_DataType")
    field_type: str

    filename: str
    file_content: bytes
    comment: str | None
    content_type: str | None

    def wd_object(self, client: WorkdayClient):
        doc = client.factory(f"ns0:{self.field_type}")
        doc.File_Content = b64encode(self.file_content).decode("utf-8")
        doc._Filename = self.filename
        doc._Content_Type = self.content_type or self._get_content_type()
        if self.comment:
            doc.Comment = self.comment
        return doc

    def _get_content_type(self):
        return mime.from_buffer(self.file_content)



from typing import Optional

from pydantic import BaseModel

# All public imports should be done through oda_wd_client.types.human_resources
__all__: list = []


class Worker(BaseModel):
    """
    Reference: https://community.workday.com/sites/default/files/file-hosting/productionapi/Human_Resources/v40.2/Get_Workers.html#Worker_DataType   # noqa
    """

    workday_id: str
    employee_number: Optional[str]
    name: str
    work_email: Optional[str]
    secondary_email: Optional[str]



from oda_wd_client.service.financial_management.api import FinancialManagement
from oda_wd_client.service.human_resources.api import HumanResources
from oda_wd_client.service.resource_management.api import ResourceManagement
from oda_wd_client.service.staffing.api import Staffing

__all__ = ["FinancialManagement", "HumanResources", "ResourceManagement", "Staffing"]



from oda_wd_client.service.staffing.types import Document


def workday_document_to_pydantic(data: dict) -> Document:
    """
    Take a dict from `suds_to_dict` of a single document and generate a Document instance instead
    """
    doc_data = data["Worker_Document_Data"]
    filename = doc_data["Filename"]
    comment = doc_data.get("Comment", "")
    employee_number = [
        field["value"]
        for field in doc_data["Worker_Reference"]["ID"]
        if field["_type"] == "Employee_ID"
    ][0]

    cat_wd = None
    cat_oda = None
    for ref in doc_data["Document_Category_Reference"]["ID"]:
        if ref["_type"] == "Document_Category__Workday_Owned__ID":
            cat_wd = ref["value"]
        elif ref["_type"] == "Document_Category_ID":
            cat_oda = ref["value"]

    return Document(
        employee_number=employee_number,
        wd_owned_category=cat_wd,
        oda_category=cat_oda,
        filename=filename,
        comment=comment,
    )



from datetime import date, datetime
from decimal import Decimal
from enum import Enum
from typing import Literal

from pydantic import BaseModel, Field

from oda_wd_client.base.types import WorkdayReferenceBaseModel

# All public imports should be done through oda_wd_client.types.financial_management
__all__: list = []


class ConversionRate(BaseModel):
    """
    Reference: https://community.workday.com/sites/default/files/file-hosting/productionapi/Financial_Management/v40.2/Put_Currency_Conversion_Rate.html#Currency_Conversion_Rate_DataType  # noqa
    """

    class RateTypeID(str, Enum):
        # Text reference to Conversion_Rate_Type in Workday
        current = "Current"
        merit = "Merit"
        budget = "Budget"
        average = "Average"

    workday_id: str | None
    # ISO 4217 defines three letters for currency ID
    from_currency_iso: str = Field(max_length=3)
    to_currency_iso: str = Field(max_length=3)

    rate: float
    rate_type_id: RateTypeID = RateTypeID.current
    effective_timestamp: datetime


class ConversionRateType(BaseModel):
    """
    Reference: https://community.workday.com/sites/default/files/file-hosting/productionapi/Financial_Management/v40.2/Put_Currency_Conversion_Rate.html#Currency_Rate_TypeObjectType  # noqa
    """

    workday_id: str
    text_id: str | None
    description: str
    is_default: bool = False


class Currency(WorkdayReferenceBaseModel):
    """
    Reference: https://community.workday.com/sites/default/files/file-hosting/productionapi/Financial_Management/v40.2/GetAll_Currencies.html#Currency_DataType  # noqa
    """

    _class_name = "CurrencyObject"
    workday_id: str = Field(max_length=3, alias="currency_code")
    workday_id_type: Literal["Currency_ID"] = "Currency_ID"
    description: str | None = None
    retired: bool = False


class Company(WorkdayReferenceBaseModel):
    """
    Reference: https://community.workday.com/sites/default/files/file-hosting/productionapi/Financial_Management/v40.2/Get_Workday_Companies.html#Company_WWS_DataType  # noqa
    """

    _class_name = "CompanyObject"
    workday_id: str
    workday_id_type: Literal["Company_Reference_ID"] = "Company_Reference_ID"
    name: str | None
    currency: Currency | None
    country_code: str | None = Field(max_length=2)


class JournalSource(WorkdayReferenceBaseModel):
    """
    Reference: https://community.workday.com/sites/default/files/file-hosting/productionapi/Financial_Management/v40.2/Submit_Accounting_Journal.html#Journal_SourceObjectType  # noqa
    """

    workday_id: str

    class JournalSourceID(str, Enum):
        # TODO: Get a new value added to workday ("Snowflake_Integration"") (Linear: DIP-1175)
        spreadsheet_upload = "Spreadsheet_Upload"

    _class_name = "Journal_SourceObject"
    workday_id_type: Literal["Journal_Source_ID"] = "Journal_Source_ID"


class LedgerType(WorkdayReferenceBaseModel):
    """
    Holds the type of ledger that we want to submit the journal to - enum values are from Workday.

    Reference: https://community.workday.com/sites/default/files/file-hosting/productionapi/Financial_Management/v40.2/Submit_Accounting_Journal.html#Ledger_TypeObjectType  # noqa
    """

    workday_id: str

    class LedgerTypeID(str, Enum):
        actuals = "Actuals"
        historic_actuals = "Historic_Actuals"

    _class_name = "Ledger_TypeObject"
    workday_id_type: Literal["Ledger_Type_ID"] = "Ledger_Type_ID"


class SpendCategory(WorkdayReferenceBaseModel):
    """
    Worktag? Seems to be dedicated field under Resource_Management at least

    Reference: https://community.workday.com/sites/default/files/file-hosting/productionapi/Resource_Management/v40.2/Submit_Supplier_Invoice.html#Spend_CategoryObjectType  # noqa
    """

    _class_name = "Spend_CategoryObject"
    workday_id: str
    workday_id_type: Literal["Spend_Category_ID"] = "Spend_Category_ID"
    name: str | None
    inactive: bool = False


class CostCenterWorktag(WorkdayReferenceBaseModel):
    """
    Reference object used as worktag in accounting. Only holds data needed for use as reference, ant not secondary
    data for cost center objects in Workday.

    Reference:  https://community.workday.com/sites/default/files/file-hosting/productionapi/Financial_Management/v40.2/Submit_Accounting_Journal.html#Audited_Accounting_WorktagObjectType  # noqa
    """

    _class_name = "Accounting_WorktagObject"
    workday_id: str
    workday_id_type: Literal["Cost_Center_Reference_ID"] = "Cost_Center_Reference_ID"
    name: str | None


class ProjectWorktag(WorkdayReferenceBaseModel):
    """
    Reference object used as worktag in accounting. Only holds data needed for use as reference, ant not secondary
    data for cost center objects in Workday.

    Reference:  https://community.workday.com/sites/default/files/file-hosting/productionapi/Financial_Management/v40.2/Submit_Accounting_Journal.html#Audited_Accounting_WorktagObjectType  # noqa
    """

    _class_name = "Accounting_WorktagObject"
    workday_id: str
    workday_id_type: Literal["Project_ID"] = "Project_ID"
    name: str | None
    inactive: bool = False


class LedgerAccount(WorkdayReferenceBaseModel):
    """
    Reference to a ledger account in Workday.

    Reference: https://community.workday.com/sites/default/files/file-hosting/productionapi/Financial_Management/v40.2/Submit_Accounting_Journal.html#Ledger_AccountObjectType  # noqa
    """

    _class_name = "Ledger_AccountObject"
    workday_id: str
    workday_id_type: Literal["Ledger_Account_ID"] = "Ledger_Account_ID"
    workday_parent_id: str
    workday_parent_type: Literal["Account_Set_ID"] = "Account_Set_ID"


class JournalEntryLineData(BaseModel):
    """
    Represents a single line in a journal entry,
    with enough information to create a journal entry in Workday.

    Reference: https://community.workday.com/sites/default/files/file-hosting/productionapi/Financial_Management/v40.2/Submit_Accounting_Journal.html#Journal_Entry_Line_DataType  # noqa
    """

    ledger_account: LedgerAccount
    debit: Decimal | None = None
    credit: Decimal | None = None
    cost_center: CostCenterWorktag | None = None
    spend_category: SpendCategory | None = None


class AccountingJournalData(BaseModel):
    """
    An accounting journal to be submitted to Workday.
    It's valid for a single company.

    Reference: https://community.workday.com/sites/default/files/file-hosting/productionapi/Financial_Management/v40.2/Submit_Accounting_Journal.html#Accounting_Journal_DataType  # noqa
    """

    accounting_date: date
    company: Company
    ledger_type: LedgerType
    journal_source: JournalSource
    journal_entry_line_data: list[JournalEntryLineData]

    @property
    def accounting_journal_id(self):
        return f"{self.accounting_date.strftime('%Y%m%d')}-{self.company.workday_id}"



from oda_wd_client.service.financial_management.types import (
    AccountingJournalData,
    Company,
    ConversionRate,
    ConversionRateType,
    CostCenterWorktag,
    Currency,
    JournalEntryLineData,
    JournalSource,
    LedgerAccount,
    LedgerType,
    ProjectWorktag,
    SpendCategory,
)

__all__ = [
    "ConversionRate",
    "Currency",
    "Company",
    "ConversionRateType",
    "JournalSource",
    "LedgerType",
    "SpendCategory",
    "CostCenterWorktag",
    "LedgerAccount",
    "JournalEntryLineData",
    "AccountingJournalData",
    "ProjectWorktag",
]



from typing import Optional, Tuple

from oda_wd_client.service.human_resources.types import Worker

WORKDAY_EMAIL_TYPES = {"WORK": "work", "HOME": "secondary"}


def _parse_worker_emails(emails: list) -> dict[str, str]:
    """
    E-mail addresses in Workday are nested deep, and we want to only extract primary addresses
    """
    ret = {}

    def _parse_address(data) -> Tuple[Optional[str], str]:
        usage_type_id = None
        addr = email["Email_Address"]
        for ud in email["Usage_Data"]:
            for td in ud["Type_Data"]:
                # Only use address if it's primary
                is_primary = td["_Primary"]
                # Lookup usage type ID - denotes if the address is work or personal
                usage_type_id = [
                    d["value"]
                    for d in td["Type_Reference"]["ID"]
                    if d["_type"] == "Communication_Usage_Type_ID"
                ]
                if is_primary and usage_type_id:
                    return usage_type_id[0], addr
        return None, ""

    for email in emails:
        wd_usage_type, _addr = _parse_address(email)
        if not wd_usage_type:
            continue
        _type = WORKDAY_EMAIL_TYPES[wd_usage_type]
        ret[_type] = _addr

    return ret


def _parse_worker_refs(refs: dict) -> Tuple[Optional[str], Optional[str]]:
    workday_id = None
    employee_number = None
    for ref in refs:
        _type = ref["_type"]
        value = ref["value"]
        if _type == "WID":
            workday_id = value
        elif _type == "Employee_ID":
            employee_number = value
    return workday_id, employee_number


def workday_worker_to_pydantic(data: dict) -> Worker:
    """
    Workday objects are painful and complex creatures, and we want to normalize them to something
    which is much easier for us to use in Python.
    """
    # Alias
    worker_data = data["Worker_Data"]
    refs = data["Worker_Reference"]["ID"]
    personal_data = worker_data["Personal_Data"]
    name_data = personal_data["Name_Data"]
    emails_data = personal_data["Contact_Data"].get("Email_Address_Data", [])

    # Lookup
    name = name_data["Legal_Name_Data"]["Name_Detail_Data"]["_Formatted_Name"]

    # Parsing
    workday_id, employee_number = _parse_worker_refs(refs)
    assert (
        workday_id
    ), "We require Workday ID for worker objects. Something is very wrong if we cannot look that up."
    emails = _parse_worker_emails(emails_data)

    return Worker(
        workday_id=workday_id,
        employee_number=employee_number,
        name=name,
        work_email=emails.get("work", None),
        secondary_email=emails.get("secondary", None),
    )



from datetime import date
from typing import Iterator

from suds import WebFault, sudsobject

from oda_wd_client.base.api import WorkdayClient
from oda_wd_client.base.tools import suds_to_dict
from oda_wd_client.service.human_resources.types import Worker
from oda_wd_client.service.human_resources.utils import workday_worker_to_pydantic


class HumanResources(WorkdayClient):
    service = "Human_Resources"

    def get_workers(
        self, as_of_date: date | None = None, return_suds_object: bool = False
    ) -> Iterator[Worker | sudsobject.Object]:
        """
        Get all workers

        Kwargs:
            as_of_date: If supplied, the Workday response will represent data that is affective at $DATE. I.e. if you
                supply a date that is two weeks in the future, the data will include all employees starting over the
                next two weeks.
            return_suds_object: If True, returns raw suds objects rather than a list of Worker instances

        """
        method = "Get_Workers"
        filters = {}
        if as_of_date:
            filters["As_Of_Effective_Date"] = as_of_date

        results = self._get_paginated(method, "Worker", filters=filters)
        for worker in results:
            yield worker if return_suds_object else workday_worker_to_pydantic(
                suds_to_dict(worker)
            )

    def _get_worker_by_id(
        self, id_: str, id_type: str, return_object: bool = False
    ) -> Worker | sudsobject.Object:
        """
        Lookup a given worker by ID
        """
        method = "Get_Workers"
        refs = self.factory("ns0:Worker_Request_ReferencesType")
        ref = self.factory("ns0:WorkerObjectType")
        obj_id = self.factory("ns0:WorkerObjectIDType")
        obj_id.value = id_
        obj_id._type = id_type
        ref.ID.append(obj_id)
        refs.Worker_Reference.append(ref)
        # Catching one class of errors to raise a more sane value error if the ID is invalid
        try:
            response = self._request(method, Request_References=refs)
        except WebFault as e:
            if e.fault.faultcode == "SOAP-ENV:Client.validationError":
                raise ValueError(f'Invalid ID of type "{id_type}"')
            raise
        worker = response.Response_Data.Worker[0]
        if return_object:
            return worker
        return workday_worker_to_pydantic(suds_to_dict(worker))

    def get_worker_by_workday_id(self, id_: str) -> Worker:
        """
        Lookup a single worker based on Workday ID
        """
        return self._get_worker_by_id(id_, "WID")

    def get_worker_by_employee_number(self, id_: str) -> Worker:
        """
        Lookup a single worker based on employee number
        """
        return self._get_worker_by_id(id_, "Employee_ID")

    def change_work_contact_info(self, *args, **kwargs) -> sudsobject.Object:
        """
        Changing contact info for a given user
        """
        method = "Change_Work_Contact_Information"
        return self._request(method, *args, **kwargs)
